/**
 * @copyright Copyright (c) Open-Xchange GmbH, Germany <info@open-xchange.com>
 * @license AGPL-3.0
 *
 * This code is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with OX App Suite. If not, see <https://www.gnu.org/licenses/agpl-3.0.txt>.
 *
 * Any use of the work other than as authorized under this license or copyright law is prohibited.
 */

const codecept = require('codeceptjs')
const querystring = require('node:querystring')

const pRetry = import('p-retry').then(module => module.default)

const cache = {}
const baseURL = codecept.config.get().helpers.Playwright.url.replace(/\/$/, '')

/**
 * Fetches data from a given URL with retry functionality.
 * @param {string} url - The URL to fetch data from.
 * @param {object} options - The options for the fetch request.
 * @returns {Promise<object>} - A promise that resolves to an object containing the fetched data and the response.
 * @throws {Error} - Throws an error if there is an HTTP request error.
 */
async function fetchWithRetry (url, options) {
  return (await pRetry)(async () => {
    const res = await fetch(url, options)
    let data
    if (res.headers.get('content-type')?.includes('application/json')) {
      data = await res.json()
    } else {
      data = await res.text()
    }
    if (data.error === 'A mail account with the given E-Mail address already exists.') {
      console.log('A mail account with the given E-Mail address already exists.')
      return true
    }
    if (data.error) throw new Error(`HTTP Request Error: ${JSON.stringify(data)}`)

    return { data, res }
  }, {
    retries: 4,
    onFailedAttempt: error => {
      console.error(`Attempt ${error.attemptNumber} failed. There are ${error.retriesLeft} retries left.`, error)
    }
  })
}

function createHttpClient (options) {
  options = options || {}
  let user = options.user || codecept.container.support('users')[0]
  if (user.toJSON) user = user.toJSON()

  async function request (url, options = {}) {
    if (user) {
      if (!cache[user.name]) await login(user)
      options.headers = { Cookie: cache[user.name].cookies, ...options.headers }
      if (!options.params) options.params = {}
      options.params.session = cache[user.name].session
    }
    if (options.params) {
      const urlObj = new URL(baseURL + url)
      urlObj.search = new URLSearchParams(options.params).toString()
      url = urlObj.toString()
    } else {
      url = baseURL + url
    }

    return fetchWithRetry(url, options)
  }

  return {
    get (url, options = {}) {
      return request(url, { ...options, method: 'GET' })
    },
    post (url, body, options = {}) {
      options = { ...options, method: 'POST', headers: { ...options.headers } }
      if (body instanceof FormData) {
        options.body = body
      } else {
        // Unify content-type header
        if (options.headers['content-type']) {
          options.headers['Content-Type'] = options.headers['content-type']
          delete options.headers['content-type']
        }
        // Set default content-type header
        if (!options.headers['Content-Type']) options.headers['Content-Type'] = 'application/json'
        options.body = JSON.stringify(body)
      }
      return request(url, options)
    },
    put (url, body, options = {}) {
      options = { ...options, method: 'PUT', headers: { 'Content-Type': 'application/json', ...options.headers } }
      if (body) options.body = JSON.stringify(body)
      return request(url, options)
    },
    delete (url, options = {}) {
      return request(url, { ...options, method: 'DELETE' })
    }
  }
}

/**
 * Logs in a user and returns the session data.
 *
 * @param {object} user - The user object containing the name and password.
 * @returns {Promise<object>} - A promise that resolves to the session data.
 * @throws {Error} - If there is an error during the login process.
 */
async function login (user) {
  if (cache[user.name]) {
    const { session, cookies } = await cache[user.name]
    try {
      if ((await fetchWithRetry(baseURL + '/api/system', {
        headers: { Cookie: cookies },
        params: {
          action: 'ping',
          timestamp: (new Date()).getTime(),
          session
        }
      })).data.data === true) {
        return cache[user.name]
      }
    } catch (e) {
      delete cache[user.name]
    }
  }

  const { data, res } = await fetchWithRetry(baseURL + '/api/login', {
    method: 'POST',
    headers: {
      'Content-Type': 'application/x-www-form-urlencoded; charset=utf-8'
    },
    body: querystring.stringify(Object.assign({
      action: 'login',
      client: 'open-xchange-appsuite'
    }, user.login || {
      name: `${user.name}${user.context.id ? '@' + user.context.id : ''}`,
      password: user.password
    }))
  })
  cache[user.name] = {
    ...data,
    cookies: res.headers.getSetCookie().map(item => item.split(';')[0]).join(';')
  }
}

module.exports = {
  createHttpClient,
  login
}
