/**
 * @copyright Copyright (c) Open-Xchange GmbH, Germany <info@open-xchange.com>
 * @license AGPL-3.0
 *
 * This code is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with OX App Suite. If not, see <https://www.gnu.org/licenses/agpl-3.0.txt>.
 *
 * Any use of the work other than as authorized under this license or copyright law is prohibited.
 */

const created = []
const users = require('../users/users')()
const util = require('../util')
const event = require('../event')
const contextService = require('../soap/services/context')
const crypto = require('node:crypto')

class Context {
  constructor ({ ctxdata, admin, auth }) {
    this.id = ctxdata.id
    this.ctxdata = ctxdata
    admin.login = admin.login || admin.name
    this.admin = admin
    this.auth = auth
    return new Proxy(this, {
      get (target, prop) {
        return prop in target ? target[prop] : target.ctxdata[prop]
      }
    })
  }

  async remove () {
    const defaultContext = await contextService.getDefault()
    // do not remove default context
    if (defaultContext !== undefined && this.ctxdata.id === defaultContext.id) throw new Error('Cannot remove default context')
    try {
      await contextService.remove(this.ctxdata.id)
    } catch (e) {
      if (!/Context \d+ does not exist/.test(e.message)) throw new util.PropagatedError(e)
      else console.error(e.message)
    }
    created.splice(0, created.length, ...created.filter(c => c !== this))
    users.splice(0, users.length, ...users.filter(u => u.context.id !== this.id))
    event.emit(event.provisioning.context.removed, this)
  }

  hasConfig (key, value) {
    // Config structure is the following { userAttributes: { entries: [{ key: 'config', value: { entries: [{ key: 'key', value: value }] }}]} }
    // Note that the soap lib will replace arrays with a single element by that element

    // check if config exists and create config path if it does not exist
    const emptyConfig = { entries: [{ key: 'config', value: null }] }
    this.ctxdata.userAttributes = this.ctxdata.userAttributes || emptyConfig
    const userAttributes = this.ctxdata.userAttributes
    const entries = userAttributes.entries = userAttributes.entries ? [].concat(userAttributes.entries) : []
    let configEntry = entries.find(e => e.key === 'config')
    if (!configEntry) {
      entries.push({ key: 'config', value: {} })
      configEntry = entries[entries.length - 1]
    }
    const config = configEntry.value = configEntry.value || { entries: [] }
    const configEntries = config.entries = config.entries ? [].concat(config.entries) : []
    let targetConfig = configEntries.find(e => e.key === key)
    if (!targetConfig) {
      configEntries.push({ key })
      targetConfig = configEntries[configEntries.length - 1]
    }
    targetConfig.value = value

    return contextService.change({ id: this.ctxdata.id, userAttributes })
  }

  hasCapability (capsToAdd) {
    return contextService.changeCapabilities(this.id, capsToAdd, undefined)
  }

  doesntHaveCapability (capsToRemove) {
    return contextService.changeCapabilities(this.id, undefined, capsToRemove)
  }

  hasAccessCombination (accessCombinationName) {
    return contextService.changeModuleAccessByName(this.id, accessCombinationName)
  }

  getModuleAccess () {
    return contextService.getModuleAccess(this.id)
  }

  hasModuleAccess (moduleAccess) {
    return contextService.changeModuleAccess(this.id, moduleAccess)
  }

  hasQuota (maxQuota) {
    this.ctxdata.maxQuota = maxQuota
    return contextService.change({ id: this.id, maxQuota })
  }

  static async create (ctx = { filestoreId: undefined, id: undefined }, adminUser = { }, auth = util.admin()) {
    adminUser = Object.assign({
      name: 'oxadmin',
      password: 'secret',
      display_name: 'context admin',
      sur_name: 'admin',
      given_name: 'context',
      email1: `${adminUser.name || 'oxadmin'}@${util.mxDomain()}`,
      primaryEmail: `${adminUser.name || 'oxadmin'}@${util.mxDomain()}`
    }, adminUser)

    const loginMappings = [`e2e-context-${crypto.randomUUID()}`]
    const newCtx = Object.assign({ maxQuota: -1, loginMappings }, ctx)
    event.emit(event.provisioning.context.create, newCtx, adminUser, auth)
    let data
    try {
      data = await contextService.create(newCtx)
    } catch (error) {
      // For mws that don't have autocontextid
      if (error.message.includes('Mandatory fields in context not set: [id]')) {
        newCtx.filestoreId = await util.getContextFilestorageId()
        newCtx.id = util.addJitter(util.userContextId())
        data = await contextService.create(newCtx)
      } else throw error
    }
    // Add loginMapping since we don't get the correct one from the mw
    data.loginMappings.push(loginMappings[0])
    const context = new Context({ ctxdata: data, admin: adminUser, auth })
    created.push(context)
    // only provide defaults for fresh contexts
    event.emit(event.provisioning.context.created, context)
    return context
  }

  static async reuse (ctx, admin = { login: 'oxadmin', password: 'secret' }, auth = util.admin()) {
    const searchCtx = Object.assign({ id: util.userContextId() }, ctx)
    const existingContext = created.find(c => c.id === searchCtx.id)
    if (existingContext) return existingContext

    const data = await contextService.get(searchCtx)

    const context = new Context({ ctxdata: data, admin, auth })
    return context
  }

  static async removeAll (auth) {
    let ctxt = created.pop()
    while (ctxt) {
      if (auth) ctxt.auth = auth
      await ctxt.remove()
      ctxt = created.pop()
    }
    return created
  }
}

module.exports = () => {
  return new Proxy(created, {
    get: function (target, prop) {
      if (prop in target) return target[prop]
      return Context[prop]
    }
  })
}
