/**
 * @copyright Copyright (c) Open-Xchange GmbH, Germany <info@open-xchange.com>
 * @license AGPL-3.0
 *
 * This code is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with OX App Suite. If not, see <https://www.gnu.org/licenses/agpl-3.0.txt>.
 *
 * Any use of the work other than as authorized under this license or copyright law is prohibited.
 */

const event = require('../event')

const users = []
const usersToRemove = []
const preprovisionedUsers = []
const util = require('../util')
const short = require('short-uuid')
const resellerUserService = require('../soap/services/resellerUser')
const oxaasService = require('../soap/services/oxaas')

class User {
  constructor (opt) {
    this.userdata = opt.user
    this.context = opt.context
    this.session = {}
    this.userAttributes = this.userdata.userAttributes || {}
    return new Proxy(this, {
      get (target, prop) {
        return prop in target ? target[prop] : target.userdata[prop]
      }
    })
  }

  async remove () {
    await resellerUserService.remove(
      this.context.id,
      this.get('id')
    ).catch(async (err) => {
      if (/No such user/.test(err.faultstring)) return // ignore
      throw new util.PropagatedError(err)
    })
    users.splice(0, users.length, ...users.filter(u => u !== this))
    event.emit(event.provisioning.user.removed, this)
  }

  async hasConfig (key, value) {
    // Config structure is the following { userAttributes: { entries: [{ key: 'config', value: { entries: [{ key: 'key', value: value }] }}]} }
    // Note that the soap lib will replace arrays with a single element by that element

    // check if config exists and create config path if it does not exist
    const userAttributes = this.userdata.userAttributes = this.userdata.userAttributes || {}
    const entries = userAttributes.entries = userAttributes.entries ? [].concat(userAttributes.entries) : []
    let configEntry = entries.find(entry => entry.key === 'config')
    if (!configEntry) {
      entries.push({ key: 'config', value: {} })
      configEntry = entries[entries.length - 1]
    }
    const config = configEntry.value
    const configEntries = config.entries = config.entries ? [].concat(config.entries) : []
    let targetConfig = configEntries.find(entry => entry.key === key)
    if (!targetConfig) {
      configEntries.push({ key })
      targetConfig = configEntries[configEntries.length - 1]
    }
    targetConfig.value = value

    return resellerUserService.change(this.context, this.userdata)
  }

  hasAccessCombination (accessCombinationName) {
    return resellerUserService.changeByModuleAccessName(this.context, this.get('id'), accessCombinationName)
  }

  getModuleAccess () {
    return resellerUserService.getModuleAccess(this.context, this.get('id'))
  }

  async hasModuleAccess (moduleAccess) {
    const currentAccess = await this.getModuleAccess()
    return resellerUserService.changeByModuleAccess(this.context, this.get('id'), currentAccess, moduleAccess)
  }

  hasQuota (quota) {
    return oxaasService.setMailQuota({
      ctxid: this.context.id,
      usrid: this.get('id'),
      quota,
      creds: util.admin()
    }).catch((err) => { throw new util.PropagatedError(err) })
  }

  hasAlias (alias) {
    // See if this alias has already been declared
    if (this.userdata.aliases.includes(alias)) {
      return
    }
    // If not, add it and save
    this.userdata.aliases.push(alias)
    const userdata = { id: this.get('id'), aliases: this.userdata.aliases }
    return resellerUserService.change(this.context, userdata)
  }

  hasCapability (capability) {
    this._setCapability(capability, true)
  }

  _setCapability (cap, value) {
    const userAttributes = this.userAttributes
    const configMap = userAttributes.entries.find(entry => entry.key === 'config') || { key: 'config' }
    configMap.value = configMap.value || { entries: [] }
    configMap.value.entries.push({ key: `com.openexchange.capability.${cap}`, value })

    return resellerUserService.change(this.context, { id: this.get('id'), userAttributes })
  }

  doesntHaveCapability (capability) {
    this._setCapability(capability, false)
  }

  get (key) {
    return this.userdata[key]
  }

  // eslint-disable-next-line camelcase
  static getRandom ({ name = 'test.user', password = util.getDefaultUserPassword(), sur_name = '', given_name = 'User' } = {}) {
    const id = short.generate().slice(0, 9).toLowerCase()
    // eslint-disable-next-line camelcase
    if (!sur_name) sur_name = id
    return {
      name: `${name}-${id}`,
      password,
      // eslint-disable-next-line camelcase
      sur_name,
      // eslint-disable-next-line camelcase
      given_name
    }
  }

  get login () {
    const { name, password } = this.userdata
    return { name, password }
  }

  static async create (user = User.getRandom(), ctx = {}, preprovisioned = false) {
    if (preprovisionedUsers.length > 0 && !preprovisioned) {
      let index = 0
      if (ctx.id) index = preprovisionedUsers.findIndex(u => u.context.id === ctx.id)
      if (index !== -1) {
        const currentUser = preprovisionedUsers[index]
        preprovisionedUsers.splice(index, 1)
        usersToRemove.push(currentUser)
        users.push(currentUser)
        return currentUser
      }
    }

    if (!ctx.id || !ctx.name) {
      const contexts = require('../contexts/reseller')()
      // @ts-ignore
      ctx = contexts[0] || await contexts.create()
    }
    const usrdata = {
      ...{
        primaryEmail: `${user.name}@${util.mxDomain()}`,
        display_name: user.name,
        sur_name: user.sur_name,
        given_name: user.given_name,
        name: user.name,
        email1: `${user.name}@${util.mxDomain()}`,
        password: user.password,
        imapLogin: user.name,
        imapServer: user.imapServer || util.imapServer(),
        smtpServer: user.smtpServer || util.smtpServer()
      },
      ...user
    }
    event.emit(event.provisioning.user.create, usrdata, ctx)

    let data
    try {
      data = await resellerUserService.createByModuleAccessName(ctx, usrdata)
    } catch (error) {
      if (!/FLD-0038/.test(error.faultstring)) {
        throw new util.PropagatedError(error)
      }
      return new Promise(function (resolve, reject) {
        setTimeout(function () {
          User.create(user, ctx).then(resolve, reject)
        }, 1000)
      })
    }
    const newUser = new User({ user: data, context: ctx })
    if (!preprovisioned) users.push(newUser)
    if (preprovisioned) preprovisionedUsers.push(newUser)
    event.emit(event.provisioning.user.created, newUser)
    return newUser
  }

  change (userdata) {
    return resellerUserService.change(this.context, userdata)
  }

  static async removeAll () {
    // Make sure to also remove the preprovisioned users
    users.splice(0, 0, ...usersToRemove.splice(0, usersToRemove.length))
    let user = users.pop()
    while (user) {
      await user.remove()
      user = users.pop()
    }
    return users
  }
}

module.exports = () => {
  return new Proxy(users, {
    // act like an array if an index is being used
    get (target, prop) {
      if (prop in target) { return target[prop] }
      return User[prop]
    }
  })
}
