import type stream from 'node:stream'
import type net from 'node:net'
import type tls from 'node:tls'
import type http from 'node:http'
import type http2 from 'node:http2'
import proxy from 'http2-proxy'
import createError from 'http-errors'
import zlib from 'zlib'

import { pipe } from './http.js'

export interface BaseProxyOptions {
  url: URL
  secure: boolean
  headers?: http.OutgoingHttpHeaders
}

export interface WebProxyOptions extends BaseProxyOptions {
  path?: string
  onReq?: proxy.http2WebOptions['onReq']
  onRes?: proxy.http2WebOptions['onRes']
}

export interface WsProxyOptions extends BaseProxyOptions {
  path: string
}

function decompress (stream: http.IncomingMessage, decompressor: zlib.Gunzip): stream.Transform {
  stream.pipe(decompressor)
  return decompressor
}

function extendRequestHeaders (reqOpts: http.RequestOptions, options: BaseProxyOptions): http.OutgoingHttpHeaders {
  const headers = reqOpts.headers ??= {}
  Object.assign(headers, options.headers)
  for (const key of Object.keys(headers)) {
    if (headers[key] === undefined || headers[key] === null) delete headers[key]
  }
  return headers
}

export function interceptData (req: http.IncomingMessage, res: http.ServerResponse, options: BaseProxyOptions): Promise<unknown> {
  const { protocol, hostname, port } = options.url
  let buffer = ''

  return new Promise((resolve, reject) => {
    proxy.web(
    // force using http2 signature of `proxy.web` for additional options
      req as unknown as http2.Http2ServerRequest,
      res as unknown as http2.Http2ServerResponse,
      {
        protocol: protocol as 'https', // needed to satisfy type checker
        hostname,
        port: parseInt(port, 10), // needed to satisfy type checker
        rejectUnauthorized: options.secure,
        onReq (_req, reqOpts) {
          const headers = extendRequestHeaders(reqOpts, options)
          headers['accept-encoding'] = 'gzip, deflate'
          return Promise.resolve() // needed to satisfy type checker
        },
        onRes (_req, _res, proxyRes_) {
        // typing bug in http2-proxy? `proxyRes` is an `IncomingMessage` at runtime
          const proxyRes = proxyRes_ as http.ServerResponse & http.IncomingMessage
          if (proxyRes.statusCode >= 400) {
            reject(createError(proxyRes.statusCode, proxyRes.statusMessage || ''))
            return Promise.resolve() // needed to satisfy type checker
          }

          const headers = proxyRes.headers
          let transform: stream.Writable
          switch (headers['content-encoding']) {
            case 'gzip': transform = decompress(proxyRes, zlib.createGunzip()); break
            case 'deflate': transform = decompress(proxyRes, zlib.createInflate()); break
            case 'br': transform = decompress(proxyRes, zlib.createBrotliDecompress()); break
            default: transform = proxyRes
          }

          transform
            .on('error', reject)
            .on('data', (buf: Buffer) => {
              buffer += buf.toString()
            })
            .on('end', () => {
              const contentType = headers['content-type']
              if (typeof contentType === 'string' && contentType.startsWith('application/json')) {
                return resolve(JSON.parse(buffer))
              }
              resolve(buffer)
            })
          return Promise.resolve() // needed to satisfy type checker
        }
      }
    )
  })
}

export async function all (req: http.IncomingMessage, res: http.ServerResponse, configs: BaseProxyOptions[]): Promise<unknown[] | undefined> {
  // TODO: property "readableEnded" is not available in type `ServerResponse` (but in `IncomingMessage`)
  if (req.aborted || (res as unknown as http.IncomingMessage).readableEnded) return

  return Promise.all(configs.map(config => interceptData(req, res, config)))
}

export async function web (req: http.IncomingMessage, res: http.ServerResponse, options: WebProxyOptions): Promise<void> {
  // TODO: property "readableEnded" is not available in type `ServerResponse` (but in `IncomingMessage`)
  if (req.aborted || (res as unknown as http.IncomingMessage).readableEnded) return

  const { protocol, hostname, port } = options.url

  return proxy.web(
    // force using http2 signature of `proxy.web` for additional options
    req as unknown as http2.Http2ServerRequest,
    res as unknown as http2.Http2ServerResponse,
    {
      protocol: protocol as 'https', // needed to satisfy type checker
      hostname,
      port: parseInt(port, 10), // needed to satisfy type checker
      path: options.path,
      // We need to increase the timeout, because the middleware might take a while to respond
      proxyTimeout: 30000,
      rejectUnauthorized: options.secure,
      async onReq (req, reqOpts, cb) {
        extendRequestHeaders(reqOpts, options)
        await options.onReq?.(req, reqOpts, cb)
      },
      async onRes (req, res, proxyRes_, cb) {
        // typing bug in http2-proxy? `proxyRes` is an `IncomingMessage` at runtime
        const proxyRes = proxyRes_ as http.ServerResponse & http.IncomingMessage
        await options.onRes?.(req, res, proxyRes, cb)
        await pipe(proxyRes, res)
      }
    }
  )
}

export async function ws (req: http.IncomingMessage, socket: net.Socket, head: Buffer, options: WsProxyOptions): Promise<void> {
  const { protocol, hostname, port } = options.url

  return proxy.ws(
    // force using http2 signature of `proxy.ws` for additional options
    req as unknown as http2.Http2ServerRequest,
    socket as tls.TLSSocket,
    head,
    {
      protocol: protocol as 'https', // needed to satisfy type checker
      hostname,
      port: parseInt(port, 10), // needed to satisfy type checker
      path: options.path,
      rejectUnauthorized: options.secure,
      onReq (_req, reqOpts) {
        extendRequestHeaders(reqOpts, options)
        return Promise.resolve() // needed to satisfy type checker
      }
    }
  )
}
