import { describe, it, expect, afterEach, vi } from 'vitest'
import vitePluginProxy from '../../dist/index'
import { createTestServer, getPort, http2Request, updateRequestManager } from '../util'
import { createServer } from 'vite'
const __dirname = import.meta.dirname
process.noDeprecation = true

describe('frontend uris', () => {
  afterEach(async () => {
    updateRequestManager()
  })

  it('works', async () => {
    const server = await createTestServer()
    server.get('/other.js', (req, res) => {
      res.send('console.log("Hello world")')
    })
    const SERVER_PORT = await getPort()
    await server.listen({ port: SERVER_PORT })

    const server1 = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        frontends: [{ target: `http://localhost:${SERVER_PORT}` }]
      })]
    })
    const PORT = await getPort()
    await server1.listen(PORT)

    const { data: index } = await http2Request(`http://localhost:${PORT}/index.js`)
    // Since vite dev server injects inline source maps, we can't use strict equality here
    expect(index).toContain('window.test = \'foo\'\n')

    const { data: other } = await http2Request(`http://localhost:${PORT}/other.js`)
    expect(other).toBe('console.log("Hello world")')

    await server1.close()
    await server.close()
  })

  it('overrides remote files with local files', async () => {
    const server = await createTestServer()
    const mockCallback = vi.fn()
    server.get('/index.js', mockCallback)
    const SERVER_PORT = await getPort()
    await server.listen({ port: SERVER_PORT })

    const server1 = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        // this has to be thought of as a remote UI with the same files
        frontends: [{ target: `http://localhost:${SERVER_PORT}` }]
      })]
    })
    const PORT = await getPort()
    await server1.listen(PORT)

    const { data } = await http2Request(`http://localhost:${PORT}/index.js`)
    // Since vite dev server injects inline source maps, we can't use strict equality here
    expect(data).toContain('window.test = \'foo\'\n')

    // give it more time to make sure, the other servers would also been called
    await new Promise(resolve => setTimeout(resolve, 50))
    expect(mockCallback).not.toHaveBeenCalled()

    await server1.close()
    await server.close()
  })

  it('overrides remote files with local files with different base configured', async () => {
    const server = await createTestServer()
    const mockCallback = vi.fn()
    server.get('/appsuite/index.js', mockCallback)
    const SERVER_PORT = await getPort()
    await server.listen({ port: SERVER_PORT })

    const server1 = await createServer({
      root: __dirname,
      logLevel: 'silent',
      base: '/appsuite/',
      plugins: [vitePluginProxy({
        // this has to be thought of as a remote UI with the same files
        frontends: [{ target: `http://localhost:${SERVER_PORT}` }]
      })]
    })
    const PORT = await getPort()
    await server1.listen(PORT)

    const { data } = await http2Request(`http://localhost:${PORT}/appsuite/index.js`)
    // Since vite dev server injects inline source maps, we can't use strict equality here
    expect(data).toContain('window.test = \'foo\'\n')

    // give it more time to make sure, the other servers would also been called
    await new Promise(resolve => setTimeout(resolve, 50))
    expect(mockCallback).not.toHaveBeenCalled()
    server1.close()
    server.close()
  })

  it('caches the handler of a file and reuses it', async () => {
    const server = await createTestServer()
    const mockCallback = vi.fn((req, res, next) => next())
    server.get('/other.js', mockCallback)
    const SERVER_PORT = await getPort()
    await server.listen({ port: SERVER_PORT })

    const server2 = await createTestServer()
    const mockCallback2 = vi.fn((req, res) => res.send('window.test = \'foo\'\n'))
    server2.get('/other.js', mockCallback2)
    const SERVER_PORT2 = await getPort()
    await server2.listen({ port: SERVER_PORT2 })

    const server1 = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        frontends: [{ target: `http://localhost:${SERVER_PORT}` }, { target: `http://localhost:${SERVER_PORT2}` }]
      })]
    })
    const PORT = await getPort()
    await server1.listen(PORT)

    const { data } = await http2Request(`http://localhost:${PORT}/other.js`)
    // Since vite dev server injects inline source maps, we can't use strict equality here
    expect(data).toContain('window.test = \'foo\'\n')
    expect(mockCallback).toHaveBeenCalledTimes(1)
    expect(mockCallback2).toHaveBeenCalledTimes(1)

    const { data: data2 } = await http2Request(`http://localhost:${PORT}/other.js`)
    expect(data2).toBe('window.test = \'foo\'\n')
    expect(mockCallback).toHaveBeenCalledTimes(1)
    expect(mockCallback2).toHaveBeenCalledTimes(2)

    await server1.close()
    await server2.close()
    await server.close()
  })
})
